__version__ = "0.0.38"
# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00__init__.ipynb.

# %% auto 0
__all__ = ['loadmod']

# %% ../nbs/00__init__.ipynb 3
from itertools import chain
from types import ModuleType
from importlib import import_module

def loadmod(mod: ModuleType):
    if hasattr(mod, '__all__'): return mod.__all__
    if hasattr(mod, '__module__'): return loadmod(import_module(mod.__module__))
    raise AttributeError(f"module {mod} has no attribute '__all__'")

# %% ../nbs/00__init__.ipynb 4
from . import (cons, type, prot, grds, enum, errs, solo, misc, spec, meta, base, mods, eggs, duck, chic)
from .cons import *
from .type import *
from .prot import *
from .enum import *
from .errs import *
from .solo import *
from .grds import *
from .misc import *
from .spec import *
# from quac.meta import *
from .base import *
# from quac.mods import *
# from quac.eggs import *
from .duck import *
from .chic import *

# %% ../nbs/00__init__.ipynb 5
__all__ = sorted(set(chain(*[
    loadmod(mod) for mod in (cons, type, prot, grds, enum, errs, solo, misc, spec, base, duck, chic)
])))
